@@ -75,7 +75,7 @@ from types import GeneratorType, ModuleType
 from collections import deque
 from binascii import a2b_hex, b2a_hex
 
-VERSION = '0.03'
+VERSION = '0.04'
 
 PYTHON3 = sys.version_info >= (3, 0)
 if PYTHON3:
@@ -162,7 +162,8 @@ except ImportError:
 
 HMAC_MOD = md5
 HMAC_SIZE = len(HMAC_MOD().digest())
-ERROR, PICKLE, CHALLENGE, RESPONSE = range(4)
+ERROR, CHALLENGE, CHALLENGE_RESPONSE, KEEPALIVE, REQUEST, RESPONSE, INIT = range(7)
+_pickle_messages = set([REQUEST, RESPONSE, INIT])
 
 DEFAULT_PORT = 11999
 
@@ -412,7 +413,14 @@ class NonblockingResponse:
     def __init__(self, client, request_id):
         self.client = client
         self.request_id = request_id
-    
+        self._have_response = False
+        self._response_val = None
+
+    @property
+    def have_response(self):
+        '''Returns True if a response is available.'''
+        return self._have_response
+
     def fileno(self):
         '''Returns the file descriptor of the associated Client instance so that
         responses can be used with select().
@@ -424,10 +432,10 @@ class NonblockingResponse:
         '''Calls read_response on the associated Client.'''
         return self.client.read_response(block)
     
-    def has_response(self):
-        '''Returns True if a response is available.'''
-        return self.client.has_response(self.request_id)
-    
+    def set_response(self, val):
+        self._response_val = val
+        self._have_response = True
+
     def wait(self, origexc=True):
         '''Waits for the request to complete.
         
@@ -436,7 +444,10 @@ class NonblockingResponse:
         exception is raised. If origexc is False, a RemoteException is raised
         instead and the remote traceback is not printed.
         '''
-        rval = self.client.wait_for_response(self.request_id)
+
+        self.client._wait_for_response(self)
+
+        rval = self._response_val
         if type(rval) is RemoteException:
             if origexc:
                 self.client.report_exception(rval)
@@ -451,11 +462,11 @@ class Connection:
         self._recv_buf = BytesIO()
         self._recv_expect = 5
         self._recv_msgtype = None
-        self._recv_hdr = None
         self._recv_data = None
         self._recv_expect_hmac = None
         self._recv_hmac = None
         self._send_hmac = None
+        self._init_data = {'version': VERSION}
         
         self._my_nonce = struct.pack('>d', time.time()) + os.urandom(HMAC_SIZE - 8)
         self._verified = False
@@ -549,9 +560,18 @@ class Connection:
         '''Returns the file descriptor of the connected socket.'''
         return self._sock.fileno()
     
-    def _got_pickle(self, p):
+    def _got_request(self, p):
         return True
 
+    def _got_response(self, p):
+        return True
+
+    def _got_init(self, p):
+        return True
+
+    def _got_keepalive(self):
+        pass
+
     def _read_data(self, block=True):
         ## attribute access protected by GIL
         sock = self._sock
@@ -570,8 +590,7 @@ class Connection:
                     return False
                 select.select([sock], [], [], None)
                 continue
-            if not data:
-                raise EOFError
+
             recv_hmac.update(data)
             recv_buf.write(data)
             newlen = self._recv_expect - len(data)
@@ -584,7 +603,10 @@ class Connection:
             recv_buf.truncate(0)
             if self._recv_msgtype is None:
                 self._recv_msgtype, self._recv_expect = struct.unpack('>BI', data)
-                continue
+                if self._recv_expect == 0:
+                    data = ''
+                else:
+                    continue
 
             if self._recv_msgtype == ERROR:
                 self._recv_msgtype = None
@@ -603,27 +625,36 @@ class Connection:
                 self._recv_expect = 5
                 self._recv_msgtype = None
                 self._recv_data = None
-                self._recv_hdr = None
                 self._recv_expect_hmac = None
 
                 if data != expect_hmac:
                     self.send_message(ERROR, b'HMAC error')
                     continue
 
-                if rmsg == PICKLE:
+                if rmsg in _pickle_messages:
                     if not self._verified:
                         self.send_message(ERROR, b'Not verified')
                         continue
 
                     p = pickle.Unpickler(BytesIO(rdata))
                     p.persistent_load = self._persistent_load
-                    return self._got_pickle(p.load())
+                    pick = p.load()
+                    if rmsg == REQUEST:
+                        return self._got_request(pick)
+                    elif rmsg == RESPONSE:
+                        return self._got_response(pick)
+                    elif rmsg == INIT:
+                        return self._got_init(pick)
+
+                elif rmsg == KEEPALIVE:
+                    self._got_keepalive()
                 elif rmsg == CHALLENGE:
-                    self.send_message(RESPONSE, rdata)
-                    self._on_ready()
-                elif rmsg == RESPONSE:
+                    self.send_message(CHALLENGE_RESPONSE, rdata)
+                elif rmsg == CHALLENGE_RESPONSE:
                     if rdata == self._my_nonce:
                         self._verified = True
+                        self.send_message(INIT, self.make_pickle(self._init_data))
+                        self._on_ready()
                     else:
                         self.send_message(ERROR, b'Invalid nonce')
                 else:
@@ -683,7 +714,7 @@ class Client(Connection):
 
         self._next_request_id = 0
         self._pending_requests = None
-        self._pending_responses = {}
+        self._nonblocking_responses = {}
         self._completion = {}
 
         self._default_thread = 0
@@ -704,7 +735,7 @@ class Client(Connection):
         self._is_ready = False
 
         self._defmod = definitions_module(definitions)
-        self._definition_count = 0
+        self._init_data['definitions_module'] = definitions
 
         if hasattr(sock_or_host, 'send'):
             self.set_socket(sock_or_host)
@@ -717,11 +748,14 @@ class Client(Connection):
     def _after_connect(self):
         self._is_ready = False
         self._pending_requests = []
+
+        source = self._defmod._source_code
+        self._definition_count = len(source)
+        self._init_data['definitions_source'] = source
+
         self._sock.setblocking(0)
         self._init_thread()
         self.send_challenge()
-        self._definition_count = -1
-        self.update_definitions()
 
     def read_response(self, block=True):
         '''Try to read a response and call its completion function (if any).
@@ -743,7 +777,7 @@ class Client(Connection):
             rq = self._pending_requests
             self._pending_requests = None
         for p in rq:
-            self.send_message(PICKLE, p)
+            self.send_message(REQUEST, p)
         
     def ready(self):
         '''Returns True if our peer is ready to accept requests.'''
@@ -762,15 +796,24 @@ class Client(Connection):
     def _thread_id(self):
         return self._default_thread
 
-    def _got_response(self, rid, response):
-        self._pending_responses[rid] = response
+    def _set_nbr_response(self, nbr, response):
+        '''Set the response value for a non-blocking response and wake it up if
+        necessary.'''
+        nbr.set_response(response)
 
-    def _got_pickle(self, p):
+    def _got_response(self, p):
         rid, response = p
         complete = self._completion.pop(rid, None)
         if not complete:
-            self._got_response(rid, response)
+            nbr = self._nonblocking_responses.pop(rid, None)
+            if nbr is None:
+                return
+            nbr = nbr()
+            if nbr is None:
+                return
+            self._set_nbr_response(nbr, response)
             return True
+
         # index 0 = oncomplete, index 1 = onerror
         compf = complete[int(type(p) is RemoteException)]
         if isinstance(compf, tuple):
@@ -846,9 +889,9 @@ class Client(Connection):
             rval = request_id
         else:
             rval = NonblockingResponse(self, request_id)
-            
+            self._nonblocking_responses[request_id] = weakref.ref(rval)
         if rq is not None:
-            self.send_message(PICKLE, rq)
+            self.send_message(REQUEST, rq)
 
         if oncomplete is None and not nonblocking:
             return rval.wait()
@@ -860,30 +903,22 @@ class Client(Connection):
         ## Don't print anything for StopIteration; remote iterators may generate this.
         if isinstance(exc.orig, StopIteration):
             return
-        
-        print('*** Remote exception from %s:' % (self._peername), file=sys.stderr)
-        print(''.join(exc.text), file=sys.stderr)
+
+        if exc.text:
+            print('*** Remote exception from %s:' % (self._peername), file=sys.stderr)
+            print(''.join(exc.text), file=sys.stderr)
 
     def stop_remote_thread(self, threadid):
         '''Stop the remote thread identified by 'threadid' once it has finished
         processing all of its requests.'''
 
-        self.send_message(PICKLE, self.make_pickle((None, None, threadid,
-                                                    None, None, None, None)))
+        self.send_message(REQUEST, self.make_pickle((None, None, threadid,
+                                                     None, None, None, None)))
 
-    def has_response(self, request_id):
-        '''Returns True if a response has been received for the given request_id.'''
-        return request_id in self._pending_responses
-    
-    def wait_for_response(self, request_id):
-        '''Wait for a specific response to complete. This method is
-        normally not called directly.'''
-
-        while request_id not in self._pending_responses:
+    def _wait_for_response(self, nbr):
+        '''Wait for a response to complete. Normally not called directly.'''
+        while not nbr._have_response:
             self.read_response()
-        rval = self._pending_responses[request_id]
-        del self._pending_responses[request_id]
-        return rval
 
     def update_definitions(self):
         '''Call this after calling define_common to send any new common
@@ -891,16 +926,8 @@ class Client(Connection):
 
         source = self._defmod._source_code
         current_count = self._definition_count
-        if current_count == len(source):
-            return
-        if current_count > 0:
-            meth = 'update_definitions'
-            arg = (source[current_count:],)
-        else:
-            meth = 'set_module'
-            arg = (self._defmod.__name__, source)
-        self.send_request(0, meth, arg, {})
-        current_count = len(source)
+        self.send_request(0, 'update_definitions', (source[current_count:],), {})
+        self._definition_count = len(source)
 
     def raw_call(self, func, args, kwargs={}, nonblocking=False,
                         oncomplete=None, onerror=None, threadid=None):
@@ -1034,8 +1061,8 @@ class Client(Connection):
         ref = RemoteProxyRef(prox, self._proxy_deleted, oid)
         self._proxy_by_id[oid] = ref
         return prox
-            
-    def close(self):
+
+    def close(self, message='Connection closed'):
         '''Close the connection and invalidate all remote proxies.'''
         Connection.close(self)
         proxies = self._proxy_by_id
@@ -1045,6 +1072,27 @@ class Client(Connection):
             if prox is not None:
                 prox._connection = _dummy_connection
 
+        ## Send an error to any outstanding requests
+        error = RemoteException(IOError(message), '')
+        comp = self._completion
+        while comp:
+            rid, (oncomplete, onerr) = comp.popitem()
+            try:
+                if isinstance(onerr, tuple):
+                    onerr[0](error, *onerr[1:])
+                else:
+                    onerr(error)
+            except Exception:
+                pass
+
+        nbrs = self._nonblocking_responses
+        while nbrs:
+            rid, nbr = nbrs.popitem()
+            nbr = nbr()
+            if nbr:
+                nbr.set_response(error)
+        
+
 class AutoThreadingClient(Client, threading.Thread):
     '''A client that automatically creates unique threads on the server for each
     thread on the client that makes remote calls.
@@ -1060,21 +1108,33 @@ class AutoThreadingClient(Client, threading.Thread):
     called. This class is also incompatible with ClientGroup.
     '''
     
-    def __init__(self, *args, **kwargs):
+    def __init__(self, key, sock_or_host=None, port=DEFAULT_PORT,
+                 definitions='easycluster_code', keep_alive_interval=10, 
+                 keep_alive_timeout=60):
         '''Creates a new AutoThreadingClient. Arguments are the same as Client.'''
         threading.Thread.__init__(self)
         self.daemon = True
 
+
         ## use pseudo-thread IDs instead of sending real ones
         self._next_thread_id = 1
         self._id_by_thread = {}
         self._response_lock = threading.Lock()
         self._have_response = threading.Condition(self._response_lock)
-        Client.__init__(self, *args, **kwargs)
+        self._keep_alive_interval = keep_alive_interval
+        self._keep_alive_timeout = keep_alive_timeout
+        self._last_keep_alive = None
+
+        Client.__init__(self, key, sock_or_host, port, definitions)
+
+        self._init_data['keep_alive_interval'] = keep_alive_interval
 
     def _init_thread(self):
         self.start()
 
+    def _got_keepalive(self):
+        self._last_keep_alive = time.time()
+
     def check_threads(self):
         '''Check for threads that have exited, and kill the 
         corresponding threads on the server.'''
@@ -1093,9 +1153,17 @@ class AutoThreadingClient(Client, threading.Thread):
         '''Read responses from the server.'''
         next_wake_time = 0
         fd = self.fileno()
+        self._got_keepalive()
+        closeargs = ()
         try:
             while True:
                 ctime = time.time()
+                timeout = self._keep_alive_timeout
+                if timeout is not None:
+                    if ctime >= self._last_keep_alive + timeout:
+                        closeargs = ('Remote server not responding',)
+                        return
+
                 if ctime >= next_wake_time:
                     next_wake_time = ctime + 1
                     # Wake up any threads that are waiting so they can check if they've
@@ -1114,7 +1182,7 @@ class AutoThreadingClient(Client, threading.Thread):
         except EOFError:
             pass
         finally:
-            self.close()
+            self.close(*closeargs)
 
     def read_response(self, *args):
         '''Do not call this method on AutoThreadingClient.'''
@@ -1130,30 +1198,30 @@ class AutoThreadingClient(Client, threading.Thread):
             self._id_by_thread[ct] = tid
         return tid
     
-    ## called with lock held
-    def _got_response(self, rid, response):
+    def _set_nbr_response(self, nbr, response):
+        with self._response_lock:
+            nbr.set_response(response)
+            self._have_response.notify_all()
+
+    def _add_pending_response(self, rid, response):
         with self._response_lock:
             self._pending_responses[rid] = response
             self._have_response.notify_all()
 
-    def wait_for_response(self, request_id):
-        '''Wait for a specific response to complete. This method is
-        normally not called directly.
+    def _wait_for_response(self, nbr):
+        '''Wait for a response to complete. This method is normally not called
+        directly.
         '''
         with self._response_lock:
-            while request_id not in self._pending_responses:
-                if not self._sock:
-                    raise IOError('Connection closed')
+            while not nbr._have_response:
                 self._have_response.wait()
 
-            rval = self._pending_responses.pop(request_id)
-        return rval
-
-    def close(self):
+    def close(self, *a):
         '''Close the connection and stop the processing thread.'''
-        ## This will shut down the socket and cause the reading thread to exit cleanly.
+        ## This will shut down the socket and cause the reading thread to exit
+        ## cleanly.
         with self._response_lock:
-            Client.close(self)
+            Client.close(self, *a)
             self._have_response.notify_all()
 
 class RootObject:
@@ -1168,10 +1236,6 @@ class RootObject:
         except _ExecReturn as e:
             return e.val
         
-    def set_module(self, name, code):
-        self._defmod = definitions_module(name, False)
-        self.update_definitions(code)
-
     def get_global(self, name):
         return _get_global(name, self._defmod, True)
 
@@ -1234,12 +1298,14 @@ class Server(Connection):
 
         # Local object that our peer has proxies for
         root = RootObject()
+        self._root = root
         self._local_object_by_id = {0: root}
         self._local_id_by_object = {id(root): 0}
         self._local_type_by_id = {}
         self._local_id_by_type = {}
         
         self._pending_responses = deque()
+        self._keep_alive_interval = None
 
         self._threads = {}
         self.set_socket(sock, peername)
@@ -1275,9 +1341,9 @@ class Server(Connection):
             rval = RemoteException(val, text)
             with self._lock:
                 rpickle = self.make_pickle((rid, rval))
-        self.send_message(PICKLE, rpickle)
+        self.send_message(RESPONSE, rpickle)
 
-    def _got_pickle(self, p):
+    def _got_request(self, p):
         rid, garbage, tid, oid, meth, args, kw = p
         if garbage:
             with self._lock:
@@ -1306,12 +1372,35 @@ class Server(Connection):
             t.queue.put((rid, oid, meth, args, kw))
         return True
 
+
+    def _got_init(self, args):
+        self._root._defmod = definitions_module(args['definitions_module'], False)
+        self._root.update_definitions(args['definitions_source'])
+        self._keep_alive_interval = args.get('keep_alive_interval')
+        if self._keep_alive_interval:
+            self._sock.setblocking(0)
+            
+        self._root._defmod.running_server = self
+        return True
+
     def run(self):
         '''Runs the main server thread.'''
         self.report_connection()
+        next_keep_alive = 0
         try:
             while True:
-                self._read_data()
+                if self._keep_alive_interval:
+                    ctime = time.time()
+                    rtime = next_keep_alive - ctime
+                    if rtime <= 0:
+                        next_keep_alive = ctime + self._keep_alive_interval
+                        self.send_message(KEEPALIVE, '')
+                        continue
+                    r, w, e = select.select([self._sock], (), (), rtime)
+                    if self._sock in r:
+                        self._read_data(False)
+                else:
+                    self._read_data(True)
         except socket.error as e:
             if e.errno != errno.ECONNRESET:
                 raise
