@@ -268,7 +268,7 @@ def do_wait_multi(rsp):
     while rsp:
         cg.read_responses(max=1)
         for r in list(rsp):
-            if r.has_response():
+            if r.have_response:
                 print('response from %d: %r' % (r.num, r.wait()))
                 rsp.discard(r)
 
@@ -343,6 +343,25 @@ def connectthread():
     active_after = l2.call('threading.active_count')
     print('Active after finishing: %d' % active_after)
 
+def test_disconnect_thread(h, amt):
+    print('Calling dosleep')
+    try:
+        h.dosleep(amt)
+    except IOError, e:
+        print('Disconnection caused error: %s' % e)
+    else:
+        assert False, 'Exception not raised on close'
+    
+
+@define_test('Test disconnection')
+def disconnect_thread():
+    t1 = threading.Thread(target=test_disconnect_thread, args=(l2, 10))
+    t1.start()
+    time.sleep(0.1)
+    l2.close()
+    time.sleep(0.2)
+    t1.join()
+
 easycluster.server_check_runner()
 
 options = OptionParser(usage='%prog [options] [host] [port]', description='Demo script for EasyCluster')
